/*
 * Description: Mini server returning the CLOCK_REALTIME clock.
 * Copyright: Catalin(ux) M. BOIE
 */

#define _GNU_SOURCE

#include <Conn.h>

/* Global variables */
static unsigned short		debug = 20;

static FILE			*Logf = NULL;
static char			*log_file = "ntime.log";
static int			port = 9000;
static short			ipv4 = 1, ipv6 = 1;
static unsigned long long	errors;


static void s_accept_error(struct Conn *C)
{
	Log(0, "%s: Cannot accept a new connection on id %llu!\n",
		Conn_getid(C));

	errors++;
}

static void s_accept(struct Conn *C)
{
	char	s[128];
	struct timespec ts;

	clock_gettime(CLOCK_REALTIME, &ts);
	snprintf(s, sizeof(s), "%ld.%ld\n", ts.tv_sec, ts.tv_nsec);
	Conn_enqueue(C, s, strlen(s));
	Conn_close(C);
}

int main(void)
{
	struct Conn *C;
	int ret;

	Logf = fopen(log_file, "w");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			log_file, strerror(errno));
		return 1;
	}
	if (debug > 0)
		setlinebuf(Logf);

	/*daemon(0, 0);*/

	Log(0, "\tPort=%d\n", port);
	Log(0, "\tLogFile=%s Debug=%d\n", log_file, debug);
	Log(0, "\tipv4=%s ipv6=%s\n", (ipv4 == 1) ? "on" : "off", (ipv6 == 1) ? "on" : "off");

	/* set library output debug */
	Conn_debug(Logf, debug);

	ret = Conn_init(0);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	if (ipv4 == 1) {
		printf("Trying to register IPv4 socket...\n");
		C = Conn_alloc();
		if (!C) {
			printf("Cannot alloc socket [%s].\n", 
				Conn_strerror());
			errors++;
		} else {
			Conn_set_socket_domain(C, PF_INET);
			Conn_set_socket_type(C, SOCK_STREAM);
			Conn_set_socket_bind_addr(C, "127.0.0.1");
			Conn_set_socket_bind_port(C, port);
			ret = Conn_commit(C);
			if (ret != 0) {
				printf("Cannot bind on ipv4 socket [%s].\n",
					Conn_strerror());
				errors++;
			}
		}
	}

	if (ipv6 == 1) {
		printf("Trying to register IPv6 socket...\n");
		C = Conn_alloc();
		if (!C) {
			printf("Cannot alloc socket [%s].\n",
				Conn_strerror());
			errors++;
		} else {
			Conn_set_socket_domain(C, PF_INET6);
			Conn_set_socket_type(C, SOCK_STREAM);
			Conn_set_socket_bind_addr(C, "::1");
			Conn_set_socket_bind_port(C, port);
			ret = Conn_commit(C);
			if (ret != 0) {
				printf("Cannot bind on ipv6 socket [%s].\n",
					Conn_strerror());
				errors++;
			}
		}
	}

	Conn_accept_cb = s_accept;
	Conn_accept_error_cb = s_accept_error;

	while (1) {
		ret = Conn_poll(-1);
		if (ret == -1) {
			Log(0, "Error calling Conn_poll [%s]!\n", Conn_strerror());
			return 1;
		} else if (ret == 0)
			break;
	}

	Conn_shutdown();

	Log(0, "Finish. Errors=%llu!\n\n",
		errors);

	fclose(Logf);

	return 0;
}
